@@ -16,6 +16,11 @@
 #include <errno.h>
 #include <vector>
 #include <list>
+#include <deque>
+
+#ifndef _REENTRANT
+#error you have to compile with -D_REENTRANT
+#endif
 
 #define QP_PRIO_DFL	0
 #define QP_PRIO_MIN	1
@@ -32,7 +37,7 @@
 #include <qpthr/qp_queue.h>
 #include <qpthr/qp_timer.h>
 #include <qpthr/qp_misc.h>
-
+#include <qpthr/qp_rwlock.h>
 
 class QpInit {
     public:
