@@ -3,7 +3,7 @@
  *	This is free software. You can redistribute it and/or modify under
  *	the terms of the GNU Library General Public License version 2.
  * 
- *	Copyright (C) 1999 by kra
+ *	Copyright (C) 1999, 2000 by kra
  *
  */
 #if ! (defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x420))
@@ -53,7 +53,6 @@ void QpRunnable::Raise()
 }
 
 
-bool      	QpThread::t_first = true;
 pthread_t 	QpThread::t_main_id;
 QpThread	*QpThread::t_main_thread;
 pthread_key_t 	QpThread::t_curr_key;
@@ -62,70 +61,66 @@ int		QpThread::t_count = 0;
 sigset_t	QpThread::t_orig_sigmask;
 
 
-QpThread::QpThread(unsigned int prio, unsigned int stack_size, const char *name, bool thr_created):
-	QpBase(name), t_mutex(name), t_was_joined(false), t_join_cond(&t_mutex, name),
-	t_cancel(false), t_sleep(false), t_sleep_cond(&t_mutex, name),
-	t_init(true), t_prio(prio), t_stack_size(stack_size), t_stack_addr(NULL),
-	t_start_mutex(name), t_except(NULL),
+QpThread::QpThread(unsigned int prio, unsigned int stack_size, const char *name):
+	QpBase(name), t_mutex(name), t_join_cond(&t_mutex, name), t_sleep_cond(&t_mutex, name),
+	t_state(T_INIT), t_sleep(false), t_cancel(false),
+	t_prio(prio), t_stack_size(stack_size), t_stack_addr(NULL),
+	t_except(NULL),
 	t_runnable(NULL)
 {
-	t_mutex.Lock();
-	if (!thr_created)
-		ThrInit(prio, stack_size);
 }
 
 QpThread::QpThread(QpRunnable *r, unsigned int prio,
 		   unsigned int stack_size, const char *name):
-	QpBase(name), t_mutex(name), t_was_joined(false), t_join_cond(&t_mutex, name),
-	t_cancel(false), t_sleep(false), t_sleep_cond(&t_mutex, name),
-	t_init(true), t_prio(prio), t_stack_size(stack_size), t_stack_addr(NULL),
-	t_start_mutex(name), t_except(NULL),
+	QpBase(name), t_mutex(name), t_join_cond(&t_mutex, name), t_sleep_cond(&t_mutex, name),
+	t_state(T_INIT), t_sleep(false), t_cancel(false),
+	t_prio(prio), t_stack_size(stack_size), t_stack_addr(NULL),
+	t_except(NULL),
 	t_runnable(r)
 {
-	t_mutex.Lock();
-	ThrInit(prio, stack_size);
+}
+
+void QpThread::ThrInitFirst(void *stack_addr)
+{
+	QPERR(pthread_key_create(&t_curr_key, NULL));
+	QPERR(pthread_sigmask(SIG_SETMASK, NULL, &t_orig_sigmask));
+	if (QpInit::Capability(QpInit::QP_SIGNAL)) {
+		t_stack_addr = (char *) stack_addr + 128;
+		QpSignal::RegisterStack(this, t_stack_addr);
+		QpSignal::ThreadSigmask();
+	}
+	t_id = t_main_id = pthread_self();
+	QPERR(pthread_setspecific(t_curr_key, this));
+	t_main_thread = this;
 }
 
 void QpThread::ThrInit(unsigned int prio, unsigned int stack_size)
 {
-	if (t_first) {
-		QPERR(pthread_key_create(&t_curr_key, NULL));
-		QPERR(pthread_sigmask(SIG_SETMASK, NULL, &t_orig_sigmask));
-		if (QpInit::Capability(QpInit::QP_SIGNAL)) {
-			t_stack_addr = &prio + 128;
-			QpSignal::RegisterStack(this, t_stack_addr);
-			QpSignal::ThreadSigmask();
-		}
-		t_id = t_main_id = pthread_self();
-		QPERR(pthread_setspecific(t_curr_key, this));
-//		t_was_joined = true;
-		t_main_thread = this;
-		t_first = false;
-		t_mutex.Unlock();
-	} else {
-		pthread_attr_t attr;
-		sched_param sch_attr;
-		QPERR(pthread_attr_init(&attr));
-		if (prio != QP_PRIO_DFL) {
-			QPERR(pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED));
-			QPERR(pthread_attr_setschedpolicy(&attr, QpInit::DflSchedPolicy()));
-			QpInit::AdjustSchedParam(prio, &sch_attr);
-			QPERR(pthread_attr_setschedparam(&attr, &sch_attr));
-		}
-		QPERR(pthread_attr_setscope(&attr, QpInit::DflScope()));
-		QPERR(pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED));
-		if ((int) stack_size >= sysconf(_SC_THREAD_STACK_MIN)) {
-#ifndef LINUX
-			QPERR(pthread_attr_setstacksize(&attr, stack_size));
-#endif
-			/*
-			 * don't allocate the stack - let the system allocate and mprotect it
-			 * in order to detect stack overflow
-			 */
-		}
-		QPERR(pthread_create(&t_id, &attr, &invoke_main, this));
-		QPERR(pthread_attr_destroy(&attr));
+	pthread_attr_t attr;
+	sched_param sch_attr;
+	QPERR(pthread_attr_init(&attr));
+	if (prio != QP_PRIO_DFL) {
+		QPERR(pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED));
+		QPERR(pthread_attr_setschedpolicy(&attr, QpInit::DflSchedPolicy()));
+		QpInit::AdjustSchedParam(prio, &sch_attr);
+		QPERR(pthread_attr_setschedparam(&attr, &sch_attr));
 	}
+	QPERR(pthread_attr_setscope(&attr, QpInit::DflScope()));
+	QPERR(pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED));
+	if ((int) stack_size >= sysconf(_SC_THREAD_STACK_MIN)) {
+#ifndef LINUX
+		QPERR(pthread_attr_setstacksize(&attr, stack_size));
+#endif
+		/*
+		 * don't allocate the stack - let the system allocate and mprotect it
+		 * in order to detect stack overflow
+		 */
+	}
+	QPERR(pthread_create(&t_id, &attr, &invoke_main, this));
+	t_count_monitor.Lock();
+	t_count++;
+	t_count_monitor.Unlock();
+	QPERR(pthread_attr_destroy(&attr));
 }
 
 QpThread::~QpThread()
@@ -135,13 +130,11 @@ QpThread::~QpThread()
 			QpSignal::UnregisterStack(this, t_stack_addr);
 		QPERR(pthread_setspecific(t_curr_key, NULL));
 		t_main_thread = NULL;
-//#ifndef LINUX
-//		/* Linux had problem with mixing exceptions and
-//		 thread specific data. */
 		pthread_key_delete(t_curr_key);
-//#endif
 	} else {
-		Join();
+		QpSLock lock(&t_mutex);
+		while (t_state == T_RUNNING)
+			t_join_cond.Wait();
 	}
 	if (t_except) {
 		delete t_except;
@@ -152,7 +145,7 @@ QpThread::~QpThread()
 void *QpThread::invoke_main(void *arg)
 {
 	QpThread *t = (QpThread *) arg;
-	t->__Begin(&t, false, true);
+	t->__Begin(&t);
 	try {
 		t->Main();
 	}
@@ -162,31 +155,22 @@ void *QpThread::invoke_main(void *arg)
 		t->t_except = e.Clone();
 	}
 	catch (...) {
-		do_exit("Unknown exception in %s::Main()\n", t->GetName() ? t->GetName() : "(null)");
+		do_exit("Unknown exception in %s::Main() (thread_id=%lu)\n",
+			t->GetName() ? t->GetName() : "(null)", t->ThreadSeq());
 	}
 	t->__Finish();
 	return NULL;
 }
 
-void QpThread::__Begin(void *stack_addr, bool set_tid, bool set_sigmask)
+void QpThread::__Begin(void *stack_addr)
 {
-	t_count_monitor.Lock();
-	t_count++;
-	t_count_monitor.Unlock();
-	if (set_tid)
-		t_id = pthread_self();
 	if (QpInit::Capability(QpInit::QP_SIGNAL)) {
 		t_stack_addr = stack_addr;
 		QpSignal::RegisterStack(this, t_stack_addr);
-		if (set_sigmask)
-			QpSignal::ThreadSigmask();
-		else
-			__OrigSigmask();
+		QpSignal::ThreadSigmask();
 	}
 	QPERR(pthread_setspecific(t_curr_key, this));
 	QPERR(pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL));
-	t_mutex.Lock();		// wait for Start()
-	t_mutex.Unlock();	// ok, leave t_mutex unlocked
 }
 
 void QpThread::__Finish()
@@ -194,17 +178,16 @@ void QpThread::__Finish()
 	if (QpInit::Capability(QpInit::QP_SIGNAL))
 		QpSignal::UnregisterStack(this, t_stack_addr);
 	
-	t_mutex.Lock();
-	t_was_joined = true;
-	t_mutex.Unlock();
-	t_join_cond.Broadcast();
+	QpSLock lock(&t_mutex);
+	t_state = T_FINISHED;
+	pthread_setspecific(t_curr_key, NULL);
 	
 	t_count_monitor.Lock();
 	t_count--;
 	t_count_monitor.Unlock();
 	t_count_monitor.Signal();
 	
-	QPERR(pthread_setspecific(t_curr_key, NULL));
+	t_join_cond.Broadcast();
 }
 
 void QpThread::__OrigSigmask()
@@ -218,26 +201,23 @@ void QpThread::Main()
 		t_runnable->InvokeMain();
 }
 
-
 void QpThread::Start()
 {
 	CheckCapability("QpThread::Start");
-	QpSLock lock(&t_start_mutex);
-	if (!t_init) {
+	QpSLock lock(&t_mutex);
+	if (t_state != T_INIT) {
+		while (t_state != T_FINISHED)
+			t_join_cond.Wait();
 		/* like in constructor */
-		Join();
 		t_cancel = false;
-		t_was_joined = false;
 		t_sleep = false;
-		t_mutex.Lock();
 		if (t_except) {
 			delete t_except;
 			t_except = NULL;
 		}
-		ThrInit(t_prio, t_stack_size);
 	}
-	t_mutex.Unlock();
-	t_init = false;
+	ThrInit(t_prio, t_stack_size);
+	t_state = T_RUNNING;
 }
 
 bool QpThread::Join(unsigned int msec, bool throw_ex)
@@ -261,9 +241,8 @@ bool QpThread::Join(const struct timespec *ts, bool throw_ex)
 }
 
 bool QpThread::JoinAbs(const struct timespec *ts, bool throw_ex) {
-	CheckCapability("QpThread::JoinAbs");
 	QpSLock lock(&t_mutex);
-	if (!t_was_joined) {
+	while (t_state != T_FINISHED) {
 		if (t_join_cond.WaitAbs(ts, false) == false) {
 			if (throw_ex)
 				throw QpTimeoutException();
@@ -351,10 +330,10 @@ bool QpThread::IsAlive()
 	QpSLock lock(&t_mutex);
 	if (t_id == t_main_id)
 		return true;
-	if (t_was_joined)
-		return false;
-	else
+	if (t_state == T_RUNNING)
 		return true;
+	else
+		return false;
 }
 
 QpThread *QpThread::Current()
@@ -445,7 +424,7 @@ void QpPThread::Main()
 			ts_abs_ptr = &ts_abs;
 		} else
 			ts_abs_ptr = NULL;
-		while (p_pool->p_job.size() == 0 && p_pool->p_running)
+		while (p_pool->p_job.empty() && p_pool->p_running)
 			if (p_pool->p_job_cond.WaitAbs(ts_abs_ptr, false) == false)
 				goto end;
 		if (!p_pool->p_running)
@@ -462,7 +441,7 @@ void QpPThread::Main()
 		p_pool->p_stat_cond.Signal();
 	}
 end:
-	p_pool->p_threads.remove(this);
+	p_pool->p_threads.erase(find(p_pool->p_threads.begin(), p_pool->p_threads.end(), this));
 	p_pool->p_threads_clean.push_back(this);
 	p_pool->p_clean_cond.Signal();
 }
@@ -484,8 +463,11 @@ QpThreadPool::QpThreadPool(unsigned int prio, unsigned int stack_size,
 			t = new QpPThread(this, prio, stack_size, life_time, name);
 			if (!t)
 				throw QpOutOfMemoryException();
-			t->Start();
+			p_mutex.Lock();
 			p_threads.push_back(t);
+			p_mutex.Unlock();
+			
+			t->Start();
 		}
 	}
 }
@@ -494,7 +476,7 @@ void QpThreadPool::Do(QpRunnable *j)
 {
 	ThreadsClean();
 
-	list<QpPThread *>::iterator li;
+	vector<QpPThread *>::iterator li;
 	unsigned int idle_threads;
 	QpPThread *t;
 
@@ -512,8 +494,10 @@ void QpThreadPool::Do(QpRunnable *j)
 		if (!t)
 			throw QpOutOfMemoryException();
 		p_mutex.Lock();
-		t->Start();
 		p_threads.push_back(t);
+		p_mutex.Unlock();
+		t->Start();
+		p_mutex.Lock();
 	}
 	p_job.push_back(j);
 	p_job_cond.Signal();
@@ -550,7 +534,7 @@ QpRunnable *QpThreadPool::WaitAbs(QpRunnable *what, const struct timespec *ts_ab
 	list<QpRunnable *>::iterator where;
 	QpSLock lock(&p_mutex);
 
-	while (p_stat.size() == 0 ||
+	while (p_stat.empty() ||
 	       (what && (where = find(p_stat.begin(), p_stat.end(), what)) == p_stat.end())) {
 		if (p_stat_cond.WaitAbs(ts_abs, false) == false) {
 			if (throw_ex)
@@ -578,7 +562,7 @@ QpThreadPool::~QpThreadPool()
 	p_running = false;
 	p_job_cond.Broadcast();
 	while (p_threads.size() || p_threads_clean.size()) {
-		while (p_threads.size() > 0 && p_threads_clean.size() == 0)
+		while (p_threads.size() > 0 && p_threads_clean.empty())
 			p_clean_cond.Wait();
 		p_mutex.Unlock();
 		ThreadsClean();
@@ -595,7 +579,7 @@ void QpThreadPool::ThreadsClean()
 	{
 	QpSLock lock(&p_mutex);
 
-	if (p_threads_clean.size() == 0)
+	if (p_threads_clean.empty())
 		return;
 	th_clean = p_threads_clean;
 	p_threads_clean.erase(p_threads_clean.begin(), p_threads_clean.end());
@@ -610,15 +594,10 @@ void QpThreadPool::ThreadsClean()
 
 QpRunnable *QpThreadPool::CurrentJob(unsigned int thr_nr)
 {
-	list<QpPThread *>::iterator li;
-
 	CheckCapability("QpThreadPool::CurrentJob");
 	QpSLock lock(&p_mutex);
 	if (thr_nr < p_threads.size()) {
-		li = p_threads.begin();
-		while (thr_nr--)
-			++li;
-		return (*li)->p_cur_job;
+		return p_threads[thr_nr]->p_cur_job;
 	} else
 		return NULL;
 }
